/***************************************************************************
 *                                                                         *
 *                  (begin: Feb 20 2003)                                   *
 *                                                                         *
 *   Parallel IQPNNI - Important Quartet Puzzle with NNI                   *
 *                                                                         *
 *   Copyright (C) 2005 by Le Sy Vinh, Bui Quang Minh, Arndt von Haeseler  *
 *   Copyright (C) 2003-2004 by Le Sy Vinh, Arndt von Haeseler             *
 *   {vinh,minh}@cs.uni-duesseldorf.de                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "iqp.h"
#include <cstring>
#include <math.h>
#include <time.h>

#include "constant.h"
#include "model.h"
#include "ali.h"
#include "utl.h"
#include "outstream.h"
#include "usertree.h"
#include "br.h"
#include "treels.h"
#include "bionj.h"
#include "clusterarr.h"
#include "rate.h"
#include "prediction.h"
#include "rannum.h"
#include "urtreepar.h"
#include "urtreepararray.h"
#include "interface.h"

// max parameter optimization iterations
int MAX_IT_ALL_PAM_OPT = 10;



using namespace std;
extern int isContinuous;

#ifdef PARALLEL
// stop signal will be send by master, set to 1 to inform that the program
// comes to an end
int stop_signal = 0;
MPI_Request stop_request = MPI_REQUEST_NULL;
#endif

//--------------------------------------------------------------------
/**
we will init all things here before starting testing this method with data
*/
void IQP::init () {
	nSeq_ = alignment.getNSeq ();
}


/**
	one step in the optimization phase (try to escape from a local optimum)
@param orderNdNoLs ramdom order sequence list, when removing/reinserting
@param tree new local optimal tree
@param bestTree best tree of the whole process so far
@param bestClusterArr cluster array of best tree
@return log likelihood of the new local optimal tree
*/
//=================================================================
double IQP::escape (Vec<int> &orderNdNoLs, UrTree &tree, UrTree &bestTree, ClusterArr &bestClusterArr) {
	//  std::cout << "escaping..." << endl;
	int *esArr_ = new int[nSeq_];
	int seqNo_;
	
	
	// generate the random array esArr_ item to 1 or 0: 1 if the related sequence to be deleted
	if (major_step_test) {
		// don't generate random, just choose to delete from 0 to probDel_*nSeq_
		for (seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++)
			if (seqNo_ < probDel_ * nSeq_)
				esArr_[seqNo_] = 1;
			else
				esArr_[seqNo_] = 0;
	} else
		for (seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++) {
			int ranNum_ = RanNum::getRanInt(100);
			esArr_[seqNo_] = ( ranNum_ < probDel_ * 100);
		}
	
	int nSeqDeled_ = 0;
	for (seqNo_ = 0; seqNo_ < nSeq_; seqNo_ ++)
		if (esArr_[seqNo_] == 1 && nSeq_) {
			
			tree.delExNd (seqNo_);
			nSeqDeled_ ++;

			if (nSeqDeled_ + 2 == nSeq_) {
				for (int count_ = seqNo_ + 1; count_ < nSeq_; count_ ++)
					esArr_[count_] = 0;
				break;
			}

		}
	
	// reinsert the deleted external nodes to the tree, using the random order from orderNdNoLs
	for (int count_ = 0; count_ < nSeq_; count_ ++) {
		seqNo_ = orderNdNoLs[count_];
		if (esArr_[seqNo_] == 1)
			tree.insertNd (seqNo_, nRep_);
	}

	delete esArr_;
	
	// create the rooted tree
	tree.reCreateRootedTree (orderNdNoLs[0]);

#ifdef PARALLEL
	if (stopSignalCome())
		return 0;
#endif

	double logLi_ = tree.reCmpLogLi (COMPLETE, bestTree, bestClusterArr, 1);
	//  double logLi_ = -INFINITIVE;
	//  double logLi_ = tree.reCmpLogLi (COMPLETE,  1);
	return logLi_;
}

//****************************************************************************************************************************************************************
/**
	optimize the parameters
*/
void IQP::optPam () {
	int isChanged_;

	UrTree bestTree_;
	bestTree_.getOptUrTree ();

#ifdef PARALLEL
	//MPI_Barrier(MPI_COMM_WORLD);
	double start_time = MPI_Wtime();
	//opt_urtree.turnOnParallelPhase(PARALLEL_LOGLI);
#else
	time_t start_time = time(NULL);
#endif
	int nIter_ = 0;
	double old_logli = 0;
	while (nIter_ < MAX_IT_ALL_PAM_OPT) {
		nIter_ ++;

		mymodel.setPamStep(nIter_);
		if (isMasterProc())
			cout << "(" << nIter_ <<") ";
		isChanged_ = 0;


		double oldGammaShape_ = myrate.getGammaShape ();
		double old_prob_invar = myrate.prob_invar_site;
		if (myrate.getType () == GAMMA && (myrate.getGammaShapeType () == ESTIMATE || myrate.getInvarSiteType() == ESTIMATE)) {
			myrate.optRatePam ();
			double newGammaShape_ = myrate.getGammaShape ();
			if (fabs (newGammaShape_ - oldGammaShape_) > EPS_GAMMA_SHAPE_ERROR ||
				fabs (myrate.prob_invar_site - old_prob_invar) > EPS_MODEL_PAM_ERROR )
				isChanged_ = 1;
		}

		if (mymodel.optPam () == 1)
			isChanged_ = 1;

		cout.precision(10);
		if (isMasterProc())
			cout << "LogL = " << opt_urtree.getLogLi() << endl << endl;

		if  (isChanged_ == 0)
			break;
		double new_logli = opt_urtree.optBranches(MAX_IT_UR_BR_PAM_OPT);
		if (fabs(old_logli - new_logli) < EPS_MODEL_FUNC_ERROR) 
			break;
		old_logli = new_logli;
	} 

#ifdef PARALLEL
	//MPI_Barrier(MPI_COMM_WORLD);
	double end_time = MPI_Wtime();
	//opt_urtree.turnOffParallelPhase();
#else
	time_t end_time = time(NULL);
#endif // PARALLEL
	if (isMasterProc())
		cout << "Time optimizing parameter(s): " << end_time - start_time << endl;

}


//****************************************************************************************************************************************************************
/**
the reestimate of parameters at the final step, based on best tree found so far
@param bestTree current best tree
@return the likelihood of the best tree
*/
double IQP::tuneFineParameter (int nIter, UrTree &bestTree) {

	bestTree.createOptUrTree();
#ifdef PARALLEL
	char filename[20];
	sprintf(filename, "tree.%i", mpi_myrank);
	//bestTree.writeNewickForm(filename);
#endif
	if (nIter > 0)
	if (mymodel.isEstedPam () == 1 ||
	        (myrate.getType () == GAMMA && (myrate.getGammaShapeType () == ESTIMATE || myrate.getInvarSiteType() == ESTIMATE)) ||
	        myrate.getType () == SITE_SPECIFIC) {

		if (isMasterProc())
			std::cout <<"Re-Optimizing parameters..." << endl;
		optPam ();
		alignment.cmpGenDis (alignment.fileName_);
	}

	double bestLogLi_ = opt_urtree.cmpLogLi ();
	if (myrate.getType () == SITE_SPECIFIC) {
		if (alignment.getNSeq() < 30 && isMasterProc()) 
			cout << endl << separator_str << endl << 
			"WARNING: Number of sequences is too small, site-specific rate estimates" << endl << 
			"         may NOT be stably inferred!" << endl << separator_str << endl;
		int nIter_ = 0;
		do {
			nIter_ ++;
			ClusterArr bestClusterArr_;
			bestTree.createCluster(bestClusterArr_);

			Vec<SeqPair> indSeqPairArr_;
			if (mymodel.ap_sitespec)
				bestTree.createAllSeqPair (indSeqPairArr_);
			else
				bestTree.createIndSeqPair (indSeqPairArr_);

			//         bestTree.draw(NON_BR_LEN, 0);
			//         OutStream::write (indSeqPairArr_, std::cout);

			if (nIter_ > 1 && isMasterProc())
				std::cout <<"Re-";

			int isChanged_ = myrate.optSpecificSite (indSeqPairArr_);
			if (isChanged_ == 0)
				break;

			if (isMasterProc())
				std::cout <<"Optimizing the tree using site-specific substitution rates... " << endl;
			bool topo_changed;
			double oldLogLi_ = bestLogLi_;
			if (nIter > 0) 
				bestLogLi_ = bestTree.doConOpt (1, topo_changed);
			else
				bestLogLi_ = bestTree.doConOpt (0, topo_changed);
			cout << "LogL = " << bestLogLi_ << endl;
			//int topDis_ = bestTree.cmpDis(bestClusterArr_) ;
			//if (topDis_ == 0)
				//break;
			if (bestLogLi_ <= oldLogLi_ + EPS_LOGLI_SITE_SPECIFIC_RATE_OPT)
				break;
		} while (nIter_ < MAX_IT_TREE_SITE_SPECIFIC_RATE_OPT);


	}

	if (nIter > 0) {
	
		if (isMasterProc())
			std::cout <<"Optimizing the final tree topology as well as branch lengths..." << endl;
		bool topo_changed;
		bestLogLi_ = bestTree.doConOpt(1, topo_changed);
	}
	std::cout.precision (10);
	if (isMasterProc())
		std::cout <<"Final best log likelihood: " << bestLogLi_ << endl;
	return bestLogLi_;
}

//****************************************************************************************************************************************************************
/**
	load all already reconstructed intermediate trees from the file into the memory
*/
void IQP::loadTreeLs (const char *treeLsFileName, int &nIter, ClusterArr &bestClusterArr, ClusterArr &allCluArr,
                      UrTree &bestTree, double &bestLogLi,
                      int &nImpTree, const char *predictionFileName,
                      double &progTime) {

	ifstream treeLsFile_;
	treeLsFile_.open (treeLsFileName);
	ofstream predictionFile_;
	predictionFile_.open (predictionFileName);

	bestLogLi = -INFINITIVE;
	nIter = 0;
	nImpTree = 0;

	while (!treeLsFile_.eof ()) {
		char aLine_[MAX_LINE_LEN];

		double tmpProgTime_;
		treeLsFile_.getline (aLine_, MAX_LINE_LEN - 1);
		char *endPos_;
		tmpProgTime_ = strtod (aLine_, &endPos_);

		double curLogLi_;
		treeLsFile_.getline (aLine_, MAX_LINE_LEN - 1);
		curLogLi_ = strtod (aLine_, &endPos_);


		Vec<char> newickTree_;
		treeLsFile_.getline (aLine_, MAX_LINE_LEN - 1);
		newickTree_.set (aLine_);

		if (newickTree_.getSize () > 0) {
			UserTree curTree_;
			curTree_.setNewickTree (newickTree_);
			int status_ = curTree_.createUrTree ();
			if (status_ == NEWICK_TREE_FORM) {
				nIter ++;
				progTime = tmpProgTime_;

				ClusterArr newClusterArr_;
				curTree_.createCluster(newClusterArr_);
				for (int count_ = 0; count_ < newClusterArr_.totalNCluster_; count_ ++)
					allCluArr.addConTree(newClusterArr_[count_]);

				if (curLogLi_ > bestLogLi + ZERO) {
					bestLogLi = curLogLi_;
					bestTree = curTree_;
					if (bestTree.cmpDis( bestClusterArr) != 0) {
						nImpTree ++;
						predictionFile_ << nIter << endl;
						bestTree.createCluster (bestClusterArr);
					}
				} //end of if curLogLi_
			} //end of if status
		}  //end of if newickTree

	}  // end of while

	treeLsFile_.close ();
	predictionFile_.close ();
}


//****************************************************************************************************************************************************************
/**
	write the parameter information, in case of restarting
*/
void IQP::writePamCheckPoint ( char  *checkPointFileName,
                               int        nIter,
                               int        max_nIter,
                               double     probDel,
                               int        nRep,
                               STOPPING_TYPE     stoppingRule,
                               int         outGrpSeqNo,
                               int         isFinished) {

	ofstream checkPointFile_;
	checkPointFile_.open (checkPointFileName);

	checkPointFile_ << "Number iterations: " << nIter << endl;
	checkPointFile_ << "Maximum number iterations: " << max_nIter << endl;
	checkPointFile_ << "Probability of deleting a sequence: " << probDel << endl;
	checkPointFile_ << "Number representatives: " << nRep - 1 << endl;
	checkPointFile_ << "Stopping rule (0: YES, 1: YES_MIN_ITER, 2: YES_MAX_ITER, 3: NO): " << stoppingRule << endl;

	checkPointFile_ << "Type of data (0:NUCLEOTIDE, 1:AMINO_ACID): " << mymodel.getDataType () << endl;


	checkPointFile_ << "Substitution model (0:HKY85, 1: TN93, 2:GTR, 3:WAG, 4:JTT, 5:VT, 6:MtREV24, 7:Blosum62, 8:Dayhoff, 9:rtREV, 10: User-defined): " << mymodel.getType () << endl;
	DVec20 baseFrqArr_;
	mymodel.getStateFrq (baseFrqArr_);
	checkPointFile_ << "Frequency of Base A: " << baseFrqArr_[BASE_A] << endl;
	checkPointFile_ << "Frequency of Base C: " << baseFrqArr_[BASE_C] << endl;
	checkPointFile_ << "Frequency of Base G: " << baseFrqArr_[BASE_G] << endl;
	checkPointFile_ << "Frequency of Base T: " << baseFrqArr_[BASE_T] << endl;
	checkPointFile_ << "Type of parameters (0:ESTIMATE,  1:USER_DEFINED, 2: EQUAL): " << mymodel.getBaseFrqType () << endl;


	checkPointFile_ << "Transition/transversion ratito: " << mymodel.getTsTvRatio () / 2.0 << endl;
	checkPointFile_ << "Type of parameters (0:ESTIMATE,  1:USER_DEFINED): " << mymodel.getTsTvRatioType () << endl;

	checkPointFile_ << "Pyridimine/purine ratito: " << mymodel.getPyPuRatio () << endl;
	checkPointFile_ << "Type of parameters (0:ESTIMATE,  1:USER_DEFINED): " << mymodel.getPyPuRatioType () << endl;

	double tsAG_, tsCT_, tvAC_, tvAT_, tvCG_, tvGT_;
	mymodel.getGenPam (tsAG_, tsCT_, tvAC_, tvAT_, tvCG_, tvGT_);

	checkPointFile_ << "Transition rate from A to G: " << tsAG_ << endl;
	checkPointFile_ << "Transition rate from C to T: " << tsCT_ << endl;

	checkPointFile_ << "Transversion rate from A to C: " << tvAC_ << endl;
	checkPointFile_ << "Transversion rate from A to T: " << tvAT_ << endl;
	checkPointFile_ << "Transversion rate from C to G: " << tvCG_ << endl;
	checkPointFile_ << "Transversion rate from G to T: " << tvGT_ << endl;
	checkPointFile_ << "Type of parameters (0:ESTIMATE,  1:USER_DEFINED): " << mymodel.getPyPuRatioType () << endl;

	RATE_TYPE rateType_ = myrate.getType ();
	checkPointFile_ << "Type of rate heterogeneity (0:UNIFORM, 1:SITE_SPECIFIC, 2:GAMMA): " << rateType_ << endl;
	checkPointFile_ << "Number rates: " << myrate.getNRate () - myrate.use_invar_site << endl;

	checkPointFile_ << "Gamma distribution parameter alpha: " << myrate.getGammaShape () << endl ;
	checkPointFile_ << "Type of parameters (0:ESTIMATE,  1:USER_DEFINED): " << myrate.getGammaShapeType () << endl;

	
	int invar_site ;
	if (!myrate.use_invar_site)
		invar_site = 0;
	else if (myrate.getInvarSiteType() == ESTIMATE)
		invar_site = 1;
	else 
		invar_site = 2;
	checkPointFile_ << "Invariant type (0: NONE, 1:ESTIMATE, 2: USER_DEFINED): " << invar_site << endl ;
	checkPointFile_ << "Propotion of invariable sites: " << myrate.prob_invar_site << endl;
	

	checkPointFile_ << "Out group sequence: " << outGrpSeqNo << endl;
	checkPointFile_ << "Finished: " << isFinished << endl;

	checkPointFile_.close ();
}

bool optimization_step = false;

void startOptimizationStep() {
	optimization_step = true;
}

void stopOptimizationStep() {
	optimization_step = false;
}

inline bool isOptimizationStep() {
	return optimization_step;
}

#ifdef PARALLEL
bool stopSignalCome() {
	if (! isOptimizationStep())
		return false;
	if (isMasterProc())
		return false;
	if (stop_request == MPI_REQUEST_NULL)
		return stop_signal;
#pragma pomp inst begin(stopsignal)
	int flag;
	MPI_Status status;
	MPI_Test(&stop_request, &flag, &status);
#pragma pomp inst end(stopsignal)
	return flag;
}
#endif

double global_bestLogLi_ = -INFINITIVE;
extern double min_negative;


//****************************************************************************************************************************************************************
/**
	main optimization step: remove nodes, reinsert by IQP, branch swapping by NNI
@param nIter number of iterations
@param probDel probability of deleting an external node
@param stoppingRule type of stopping rule: YES, YES_MIN_ITER, YES_MAX_ITER, NO \
	YES_MIN_ITER: number of iterations is at least the user-predefined number
	YES_MAX_ITER: number of iterations is at most the user-predefined number 
@param outGrpSeqNo outgroup sequence number
@param bestTree current best tree, to return
@param conTree consensus Tree
@param beginTime beginning time
@param progTime running time
@return likelihood of the best tree
*/
double IQP::escapeLocalOpt (int &nIter, int max_nIter, double probDel, int nRep, STOPPING_TYPE stoppingRule, UrTree &bestTree, const char *userTreeFile, int outGrpSeqNo, Vec<char> &conTree,
                            time_t &beginTime, double &progTime) {
	char predictionFileName_[FILENAME_LEN];
	char treeLsFileName_[FILENAME_LEN];

	char bionjFileName_[FILENAME_LEN];
	char distFileName_[FILENAME_LEN];
	char checkPointFileName_[FILENAME_LEN];

	strcpy (predictionFileName_, alignment.fileName_);
	strcpy (treeLsFileName_, alignment.fileName_);
	strcpy (bionjFileName_, alignment.fileName_);
	strcpy (distFileName_, alignment.fileName_);
	strcpy (checkPointFileName_, alignment.fileName_);

	char *predictionTail_ = ".iqpnni.prediction";
	char *treeLsTail_ = ".iqpnni.treels";
	char *bionjTail_ = ".iqpnni.bionj";
	char *distTail_ = ".iqpnni.dist";

	char *checkPointTail_ = ".iqpnni.checkpoint";

	strcat (predictionFileName_, predictionTail_);
	strcat (treeLsFileName_, treeLsTail_);
	strcat (bionjFileName_, bionjTail_);
	strcat (distFileName_, distTail_);
	strcat (checkPointFileName_, checkPointTail_);

	char topoFileName_[FILENAME_LEN];
	strcpy (topoFileName_, alignment.fileName_);
	char *topoTail_ = ".iqpnni.treefile";
	strcat (topoFileName_, topoTail_);
	
	nSeq_ = alignment.getNSeq ();
	nRep_ = nRep;
	probDel_ = probDel;

	int nSite_ = alignment.getNSite ();
	if (isMasterProc())
		std::cout <<"number sequences = " << nSeq_ << "; sequence length = " << nSite_ << endl;

	ClusterArr allCluArr_;
	allCluArr_.setupIndex ();

	int nPredictedIter_ = MAX_ITERATION;
	int nImpTree_ = 1;
	int iterNo_ = 1;

	Prediction prediction_;

	double stoppedProgTime_ = 0.0;
	int maxNIter_ = nIter;

	if (isContinuous == 0) {
#ifdef PARALLEL
		double start_time = MPI_Wtime();
#else
		time_t start_time = time(NULL);
#endif // PARALLEL
		
		UserTree initialTree_;
		bool hasInitTree = false;
		
		if (userTreeFile != NULL) {
			std::ifstream userTreeFile_;
			userTreeFile_.open (userTreeFile);
			if (userTreeFile_ != 0) {
				initialTree_.readFile (userTreeFile);
				initialTree_.createUrTree ();
				hasInitTree = true;
			} 
			userTreeFile_.close();
		}
		
		if (!hasInitTree) {
			if (isMasterProc()) {
				cout << "Creating neighbour joining tree... ";
				cout.flush();
				// only master does the neighbour joining tree
				BioNj bioNj_;
				bioNj_.create (distFileName_, bionjFileName_);
			}
			if (isMasterProc()) {
#ifdef PARALLEL
				cout << MPI_Wtime() - start_time << " s" << endl;
#else
				cout << time(NULL) - start_time << " s" << endl;
#endif // PARALLEL
			}
#ifdef PARALLEL
			// wait until master finishes doing the neighbor joining tree
			MPI_Barrier(MPI_COMM_WORLD);
#endif // PARALLEL
			initialTree_.readFile (bionjFileName_);
	
	
			initialTree_.createUrTree ();
			
		} else {
			if (isMasterProc())
				cout  << "Start the search from the user tree!" << endl;
		}
		//   bionjTree_.draw (BR_LEN, 0);

		// check the tree, positive branch lengths...
		initialTree_.checkUrTree ();


		// optimizing the branch length
		if (isMasterProc()) {
			cout << "Optimizing branch lengths... ";
			cout.flush();
		}
#ifdef PARALLEL
		start_time = MPI_Wtime();
#else
		start_time = time(NULL);
#endif // PARALLEL
		bool topo_changed;
		double bionjLogLi_ =  initialTree_.doConOpt (0, topo_changed, 1);
		//cout << "min prob=" << min_negative << endl;
		if (isMasterProc()) {
#ifdef PARALLEL
			cout << MPI_Wtime() - start_time << " s" << endl;
#else
			cout << time(NULL) - start_time << " s" << endl;
#endif // PARALLEL

		}
		//  std::cout << bionjLogLi_ << endl;

		// estimate the parameters
		if ( mymodel.isEstedPam () == 1 ||
		        (myrate.getType () == GAMMA && (myrate.getGammaShapeType () == ESTIMATE ||
					myrate.getInvarSiteType() == ESTIMATE)) ) {
			//cout << " optPam: " << mpi_myrank << endl;
			optPam ();
			if (program_mode == PROG_PARM_ESTIMATE)
				Finalize(0);
			alignment.cmpGenDis (alignment.fileName_);
		}

		if (program_mode != PROG_ALL)
			Finalize(0);
		// do the NNI
		if (isMasterProc()) {
			if (maxNIter_ > 0)
				cout << "Doing Nearest Neighbour Interchange... ";
			else
				cout << "Reoptimizing branch lengths... ";
			cout.flush();
		}
#ifdef PARALLEL
		start_time = MPI_Wtime();
#else
		start_time = time(NULL);
#endif // PARALLEL
		if (maxNIter_ > 0)
			bionjLogLi_ = initialTree_.doConOpt (1, topo_changed);
		else
			bionjLogLi_ = initialTree_.doConOpt (0, topo_changed);

		if (isMasterProc()) {
#ifdef PARALLEL
			cout << MPI_Wtime() - start_time << " s" << endl;
#else
			cout << time(NULL) - start_time << " s" << endl;
#endif // PARALLEL

		}

		if (nni_test) 
			Finalize(0);
		
		bestTree = initialTree_;
		global_bestLogLi_ = bionjLogLi_;
		bestTree.createCluster (bestClusterArr_);

		for (int clusterNo_ = 0; clusterNo_ < bestClusterArr_.totalNCluster_; clusterNo_ ++)
			allCluArr_.addConTree(bestClusterArr_[clusterNo_]);

		time_t currentTime_;
		time(&currentTime_);

		//cout << mpi_myrank << ": we are here!" << endl;


		progTime = difftime (currentTime_, beginTime);

		if (isMasterProc()) {
			bestTree.writeNewickForm (treeLsFileName_, 0,  progTime);

			std::cout << "We have constructed the initial tree !!!" << endl;
			std::cout.precision (10);
			std::cout << "The currently best log likelihood = " << global_bestLogLi_  << endl;

			std::ofstream predictionFile_;
			predictionFile_.open (predictionFileName_);
			predictionFile_ << 1 << endl;
			predictionFile_.close ();
			bestTree.writeTop(topoFileName_, 0);
		}
		int isFinished_ = 0;
		if (isMasterProc())
			writePamCheckPoint(checkPointFileName_,
			                   nIter, max_nIter, probDel, nRep, stoppingRule, outGrpSeqNo,
			                   isFinished_);

	} else {//if isContinuous == 0
		if (isMasterProc())
			std::cout <<"Loading the data from the last stopped point ...";
		loadTreeLs (treeLsFileName_, iterNo_, bestClusterArr_, allCluArr_, bestTree, global_bestLogLi_, nImpTree_, predictionFileName_, stoppedProgTime_);

		if (nImpTree_ >= START_PREDICTION /*&& stoppingRule != NO*/) {
			double tmpNPreIter_;
			prediction_.predict(predictionFileName_, 0.95, tmpNPreIter_);
			nPredictedIter_ = static_cast<int> (tmpNPreIter_);
			if (nPredictedIter_ <= iterNo_)
				nPredictedIter_ = MAX_ITERATION;
			else {
				if (stoppingRule != NO) {
					maxNIter_ = Utl::getMax (nIter, nPredictedIter_);
					if (stoppingRule == YES_MAX_ITER)
						maxNIter_ = Utl::getMin (max_nIter, maxNIter_);
				}
	/*
				if (stoppingRule == YES)
					maxNIter_ = Utl::getMax (nIter, nPredictedIter_);
				if (stoppingRule == YES_MIN_ITER)
					maxNIter_ = Utl::getMax (nIter, nPredictedIter_);
				if (stoppingRule == YES_MAX_ITER)
					maxNIter_ = Utl::getMin (nIter, nPredictedIter_);
	*/
			}
		}

		if (isMasterProc())
			std::cout << endl << "Continuing from the last stopped point ..." << endl;
	}

	double lastProgTime_ = 0.0;

	//***********************************************************************************************
	//  start IQPNNI
	//***********************************************************************************************

#pragma pomp inst begin(iqp)

#ifdef PARALLEL

	double start_time = MPI_Wtime();

	// tree_from_master is used for receiving data from master, non blocking communication
	UrTreePar tree_from_master(bestTree);

	// tree array of master, use to receive data from all slaves
	UrTreeParArray master_tree_array(bestTree);
	if (isMasterProc())
		master_tree_array.receiveFromSlaves();
	else
		// start receiving from master
		tree_from_master.receiveFromMaster();

	// array of tree, used to broadcast tree to all slaves

	UrTreeParArray master_bcast;

	// tree_to_master is used for sending data to master, non blocking communication
	UrTreePar tree_to_master;

	int TAG_STOP_SIGNAL = 100;
	if (isSlaveProc())
		MPI_Irecv(&stop_signal, 1, MPI_INT, mpi_master_rank, TAG_STOP_SIGNAL, MPI_COMM_WORLD, &stop_request);

		
#endif // PARALLEL

	startOptimizationStep();
#ifdef PARALLEL
	//int process_index[mpi_size];
	//int process_count = 0;
#endif // PARALLEL
	
	while (iterNo_ < maxNIter_) {
		/*
		if (isMasterProc()) {
			cout << "Optimization step " << iterNo_ << "  ";
			cout.flush();
		}
		*/
		iterNo_ ++;

#ifdef PARALLEL
		if (isSlaveProc()) {
			if (stopSignalCome())
				break;

			while (tree_from_master.isDataArrived()) {
				tree_from_master.unpackData();
				cout.precision(10);
				//cout << "process " << mpi_myrank << " received from master with logli=" << tree_from_master.getLogLi();
				if (tree_from_master.getLogLi() > bestTree.getLogLi()) {
					//cout << " updated!" << endl;
					bestTree.clone(tree_from_master);
					global_bestLogLi_ = tree_from_master.getLogLi();
					bestTree.createCluster (bestClusterArr_);
					/*
					for (int count_ = 0; count_ < bestClusterArr_.totalNCluster_; count_ ++)
						allCluArr_.addConTree(bestClusterArr_[count_]);
					*/

				} else {
					//cout << " ...not better!" << endl;
				}
				
				bool can_receive = tree_from_master.isReadyToReceive();
				// start a new receive
				tree_from_master.receiveFromMaster();
				if (can_receive) // wait until receive
					tree_from_master.waitReceive();
			}
		}
#endif // PARALLEL

		UrTree curTree_;
		double curLogLi_;

#ifdef PARALLEL		
		// process only useful for master, it is the process rank which send tree to master
		int process = 0;
#endif

		if (
#ifdef PARALLEL 
			isSlaveProc() || mpi_size == 1
#else
			true
#endif // PARALLEL
		) {
			// if i'm a slave proc, or only 1 process spawn
			Vec<int> orderNdNoLs_;
			RanNum::createOrderNd (nSeq_, orderNdNoLs_);
			//cout << " Process " << mpi_myrank << ": " ;
			//orderNdNoLs_.writeItem(cout);
			//cout << endl;
			
			curTree_ = bestTree;
			curLogLi_ = escape (orderNdNoLs_, curTree_, bestTree, bestClusterArr_);
			
			//curTree_.writeNewickForm ("test.tree");
			//Finalize(0);
#ifdef PARALLEL
			// non blocking send the result tree to master

			if (stopSignalCome())
				break;
			if (mpi_size > 1) {
				tree_to_master.clone(curTree_);
				tree_to_master.sendToMaster();
			}
			//cout << mpi_myrank<< ": hi!" << endl;
#endif // PARALLEL

		} else {
			// master work: receive tree from all slaves
			//cout << mpi_myrank<< "->: wait for slaves " << endl;
#ifdef PARALLEL
			
			process = master_tree_array.waitForSlaves();
			//cout << mpi_myrank<< ": receive from " << process << endl;
			master_tree_array.getItem(process)->unpackData();
			//cout << "Process " << process << " sends a tree" << endl;
			curTree_.clone((UrTree*)master_tree_array.getItem(process));
			curLogLi_ = curTree_.getLogLi();

			// if still to run, post receive from this process
			//if (iterNo_ + master_tree_array.runningSlaves() < maxNIter_)
			master_tree_array.getItem(process)->receiveFromProcess(process);
#endif

		}

		ClusterArr newClusterArr_;
		curTree_.createCluster(newClusterArr_);
		for (int count_ = 0; count_ < newClusterArr_.totalNCluster_; count_ ++)
			allCluArr_.addConTree(newClusterArr_[count_]);


		time_t currentTime_;
		time(&currentTime_);
		progTime = difftime (currentTime_, beginTime) + stoppedProgTime_;

		int nHour_ = static_cast<int> (progTime / 3600);
		int nMin_ = static_cast<int> ( (progTime - nHour_ * 3600) / 60);
		int nSec_ = static_cast<int> (progTime - nMin_ * 60 - nHour_ * 3600);

		if (isMasterProc())
			curTree_.writeNewickForm (treeLsFileName_, 1, progTime);



		int remNHour_ = 0;
		int remNMin_ = 0;
		int remNSec_ = 1;
		if (maxNIter_ > iterNo_) {
			int remTime_ = static_cast<int> ( (progTime / iterNo_) * (maxNIter_ - iterNo_));
/*
#ifdef PARALLEL
			if (mpi_size > 1)
				remTime_ = static_cast<int>(remTime_ / (mpi_size - 1));
#endif
*/

			remNHour_ = static_cast<int> (remTime_ / 3600);
			remNMin_ = static_cast<int> ( (remTime_ - remNHour_ * 3600) / 60);
			remNSec_ = static_cast<int> (remTime_ - remNHour_ * 3600 - remNMin_ * 60);
		}

		if (isMasterProc())
			if (progTime -  lastProgTime_ > 10.0 || curLogLi_ > global_bestLogLi_ || iterNo_ >= maxNIter_) {
				std::cout << iterNo_ << " Iterations / time elapsed = " << nHour_ <<"h:" << nMin_<<"m:" << nSec_<<"s";
				if (maxNIter_ > iterNo_)
					std::cout << "  (will finish in " << remNHour_<<"h:" << remNMin_ <<"m:" << remNSec_ <<"s)" << endl;
				else
					cout << endl;
				lastProgTime_ = progTime;
			}

		
		if (curLogLi_ > global_bestLogLi_) {
			if (curTree_.cmpDis( bestClusterArr_) != 0) {
				if (isMasterProc()) {
					// start the prediction
					ofstream predictionFile_;
					predictionFile_.open (predictionFileName_, ios::app);
					predictionFile_ << iterNo_ << endl;
					predictionFile_.close ();

					nImpTree_ ++;
					if (nImpTree_ >= START_PREDICTION /*&& stoppingRule != NO*/) {
						double tmpNPreIter_;
						prediction_.predict(predictionFileName_, 0.95, tmpNPreIter_);
						if (static_cast<int> (tmpNPreIter_) >
						        iterNo_) {
							nPredictedIter_ = static_cast<int> (tmpNPreIter_);
							if (stoppingRule != NO) {
								maxNIter_ = Utl::getMax (nIter, nPredictedIter_);
								if (stoppingRule == YES_MAX_ITER)
									maxNIter_ = Utl::getMin (max_nIter, maxNIter_);
							}
/*
							if (stoppingRule == YES)
								maxNIter_ = Utl::getMax (nIter, nPredictedIter_);
							if (stoppingRule == YES_MIN_ITER)
								maxNIter_ = Utl::getMax (nIter, nPredictedIter_);
							if (stoppingRule == YES_MAX_ITER)
								maxNIter_ = Utl::getMin (nIter, nPredictedIter_);
							if (maxNIter_ > max_nIter) maxNIter_ = max_nIter;
*/
						}  // end of if tmpNPreIter_ > iterNo
					}
					
					curTree_.writeTop(topoFileName_, 0);

				} // if (isMasterProc())

#ifdef PARALLEL
				// broadcast the better to slaves
				if (isMasterProc())
					master_bcast.broadcastToSlaves(curTree_, process);
#endif
				global_bestLogLi_ = curLogLi_;
				bestTree = curTree_;
				bestTree.createCluster (bestClusterArr_);

				if (isMasterProc()) {
					std::cout.precision (10);

					std::cout <<"GOOD NEWS: BETTER TREE FOUND: THE CURRENTLY BEST LOG LIKELIHOOD = " << global_bestLogLi_ << endl;
					if (nImpTree_ >= START_PREDICTION /*&& nPredictedIter_ < MAX_ITERATION*/) {
						double neededTime_ = ( progTime / iterNo_) * (nPredictedIter_ - iterNo_);
						int nNeededHour_ = static_cast<int> (neededTime_ / 3600);
						int nNeededMin_ = static_cast<int> ((neededTime_ - nNeededHour_ * 3600)/ 60);
						int nNeededSec_ = static_cast<int> (neededTime_ - nNeededMin_ * 60 - nNeededHour_ * 3600);
						std::cout <<"PREDICTION: " << (nPredictedIter_ - iterNo_) << 
						" iterations more are needed (about " << nNeededHour_ <<"h:" << nNeededMin_ <<"m:"<<nNeededSec_<<"s)" <<endl;
					}
				}
			}
		}

		//    char answer_;
		if (iterNo_ >= maxNIter_) {
			if (isMasterProc() && nPredictedIter_ < MAX_ITERATION) {
				if (iterNo_ >= nPredictedIter_) 
					std::cout << endl << "WE HAVE FOUND THE BEST TREE WITH 95% CONFIDENCE" << endl;
				else
					cout << endl << separator_str << endl <<
					"WARNING: According to the stopping rule, at least " << (nPredictedIter_ - iterNo_) << 
					" more iterations are" << endl << 
					"         suggested to ensure that with a 95% confidence the current search" << endl <<
					"         will not detect a better tree. If you want to specify the confidence" << endl << 
					"         level, you should switch on the stopping rule in the option menu." << endl << 
					separator_str << endl;
			}
			break;
		}

		if (iterNo_ >= nIter && nPredictedIter_ == MAX_ITERATION) {
			if (isMasterProc())
				std::cout << "WE HAVE CREATED " << iterNo_ << " TREES" << endl;
			break;
		}
	} //end of while


#ifdef PARALLEL
	// now cancel all non blocking communication
	if (isMasterProc()) {
		master_tree_array.cancelCommunications();
		master_bcast.cancelCommunications();
	} else {
		tree_from_master.cancelCommunications();
		tree_to_master.cancelCommunications();
	}

	MPI_Request *request = new MPI_Request[mpi_size];
	int *stop = new int[mpi_size];

	// tell all slaves to stop
	if (isMasterProc()) {
		for (int cnt = 0; cnt < mpi_size; cnt++)
			if (cnt != mpi_master_rank) {
				stop[cnt] = 1;
				MPI_Isend(&stop[cnt], 1, MPI_INT, cnt, TAG_STOP_SIGNAL, MPI_COMM_WORLD,
				          &request[cnt]);
			}
	} else { // slave
		if (stop_request != MPI_REQUEST_NULL)
			MPI_Cancel(&stop_request);
	}

	if (isMasterProc()) {
		cout << "Time optimization step: " << MPI_Wtime() - start_time << endl;
	}
	
	delete stop;
	delete request;
	
	if (isMasterProc()) {
		cout << "Wait for all slaves to stop... ";
		cout.flush();
		start_time = MPI_Wtime();
	}
	MPI_Barrier(MPI_COMM_WORLD);
	if (isMasterProc())
		cout << MPI_Wtime() - start_time << " s" << endl;

	// broadcast the best tree to slaves
	if (isMasterProc())
		master_bcast.broadcastToSlaves(bestTree, mpi_master_rank);
	else {
		tree_from_master.receiveFromMaster();
		tree_from_master.waitReceive();
		tree_from_master.unpackData();
		bestTree.clone(tree_from_master);
		bestTree.createCluster (bestClusterArr_);
	}

	//MPI_Barrier(MPI_COMM_WORLD);


#endif // PARALLEL

#pragma pomp inst end(iqp)


	stopOptimizationStep();

	global_bestLogLi_ = tuneFineParameter (nIter, bestTree);
	nIter = iterNo_;

	if (isMasterProc()) {
		std::cout <<"Constructing the majority rule consensus tree..." << endl;
		allCluArr_.createConTree(nIter, conTree);
	}
	int isFinished_ = 1;
	if (isMasterProc())
		writePamCheckPoint(checkPointFileName_,
		                   nIter, max_nIter, probDel, nRep, stoppingRule, outGrpSeqNo, isFinished_);
	time_t currentTime_;
	time(&currentTime_);
	progTime = difftime (currentTime_, beginTime) + stoppedProgTime_;
	return global_bestLogLi_;
}
